from typing import Dict
import json
from pathlib import Path

import netCDF4 as nc
import numpy as np


class UgridReader:
    def __init__(self, file_path: Path) -> None:
        self._ncfile_path = file_path

        if not self._ncfile_path.exists():
            raise OSError(f'File "{self._ncfile_path}" not found.')

        self._explorer = NCExplorer(self._ncfile_path)

    def read_mesh1d_network1d(self, mesh1d: "Mesh1d") -> None:
        """
        Read Ugrid from netcdf and return dflowfm cstructure with grid
        """
        # If the mesh is not given (default), use the networks one
        ds = nc.Dataset(self._ncfile_path)  # type: ignore[import]

        # Read mesh1d
        for meshkey, nckey in self._explorer.mesh1d_var_dict.items():
            setattr(mesh1d, meshkey, self._read_nc_attribute(ds[nckey]))

        # Read network variables
        for meshkey, nckey in self._explorer.network1d_var_dict.items():
            setattr(mesh1d, meshkey, self._read_nc_attribute(ds[nckey]))

        # Process network
        mesh1d._process_network1d()

        # self.network.mesh1d.add_from_other(mesh1d)

        ds.close()

    def read_mesh2d(self, mesh2d: "Mesh2d") -> None:

        ds = nc.Dataset(self._ncfile_path)  # type: ignore[import]

        # Read mesh1d
        for meshkey, nckey in self._explorer.mesh2d_var_dict.items():
            setattr(mesh2d, meshkey, self._read_nc_attribute(ds[nckey]))

        ds.close()

    def read_link1d2d(self, link1d2d: "Link1d2d") -> None:
        ds = nc.Dataset(self._ncfile_path)  # type: ignore[import]

        # Read mesh1d
        for meshkey, nckey in self._explorer.link1d2d_var_dict.items():
            setattr(link1d2d, meshkey, self._read_nc_attribute(ds[nckey]))

        ds.close()

    def _read_nc_attribute(self, attr: nc._netCDF4.Variable) -> np.ndarray:
        """Read values from netcdf attribute.
        - Character arrays are converted to strings
        - Masked arrays are converted to regular arrays where the masked values are filled with the fillvalue
        - The numerical arrays are subtracted by the start value. This is often 1, while python is 0-based.

        Args:
            attr (netCDF4._netCDF4.Variable): Attribute in the file

        Returns:
            np.ndarray: returned array
        """
        values = attr[:]
        if values.dtype == "S1":
            # Convert to strings
            arr = np.array(list(map(str.strip, nc.chartostring(values.data))))  # type: ignore[import]

        else:
            # Get data from masked array
            arr = values[:].data
            # Fill masked values with fillvalue
            if values.mask.any():
                arr[values.mask] = attr._FillValue
            # Substract start index
            if hasattr(attr, "start_index"):
                arr -= attr.start_index

        return arr


class NCExplorer:
    def __init__(self, file_path: Path):

        self.ds = nc.Dataset(file_path)  # type: ignore[import]

        # Read the key conventions from json, based on nc version number
        with open(Path(__file__).parent.joinpath("ugrid_conventions.json"), "r") as f:
            variable_names = json.load(f)

        self.network1d_var_dict = variable_names["network1d"]
        self.mesh1d_var_dict = variable_names["mesh1d"]
        self.mesh2d_var_dict = variable_names["mesh2d"]
        self.link1d2d_var_dict = variable_names["link1d2d"]

        self.network1d_key = None
        self.mesh1d_key = None
        self.mesh2d_key = None
        self.link1d2d_key = None

        self._get_mesh1d_network1d_key()
        self._get_mesh2d_key()
        self._get_link1d2d_key()

        self._update_dicts()

        self.ds.close()

    def _get_mesh1d_network1d_key(self) -> None:
        for var_key, ncdata in self.ds.variables.items():
            attributes = ncdata.ncattrs()
            if (
                (
                    ("cf_role" in attributes)
                    and (ncdata.getncattr("cf_role") == "mesh_topology")
                )
                and (
                    ("topology_dimension" in attributes)
                    and (ncdata.getncattr("topology_dimension") == 1)
                )
                and ("coordinate_space" in attributes)
            ):
                self.mesh1d_key = var_key
                self.network1d_key = ncdata.getncattr("coordinate_space")
                break

    def _get_mesh2d_key(self) -> None:
        for var_key, ncdata in self.ds.variables.items():
            attributes = ncdata.ncattrs()
            if (
                ("cf_role" in attributes)
                and (ncdata.getncattr("cf_role") == "mesh_topology")
            ) and (
                ("topology_dimension" in attributes)
                and (ncdata.getncattr("topology_dimension") == 2)
            ):
                self.mesh2d_key = var_key
                break

    def _get_link1d2d_key(self) -> None:
        for var_key, ncdata in self.ds.variables.items():
            attributes = ncdata.ncattrs()
            if ("cf_role" in attributes) and (
                ncdata.getncattr("cf_role") == "mesh_topology_contact"
            ):
                self.link1d2d_key = var_key
                break

    def _update_dicts(self):
        """Prepend the key to each dictionary value
        Check existance of values from dictionary in netcdf"""
        # Add deduced keys to dictionary
        if self.network1d_key is not None:
            self.network1d_var_dict = {
                k: [self.network1d_key + v for v in vs]
                for k, vs in self.network1d_var_dict.items()
            }
            self._check_existence(self.network1d_var_dict)

        if self.mesh1d_key is not None:
            self.mesh1d_var_dict = {
                k: [self.mesh1d_key + v for v in vs]
                for k, vs in self.mesh1d_var_dict.items()
            }
            self._check_existence(self.mesh1d_var_dict)

        if self.mesh2d_key is not None:
            self.mesh2d_var_dict = {
                k: [self.mesh2d_key + v for v in vs]
                for k, vs in self.mesh2d_var_dict.items()
            }
            self._check_existence(self.mesh2d_var_dict)

        if self.link1d2d_key is not None:
            self.link1d2d_var_dict = {
                k: [self.link1d2d_key + v for v in vs]
                for k, vs in self.link1d2d_var_dict.items()
            }
            self._check_existence(self.link1d2d_var_dict)

    def _check_existence(self, dct: Dict) -> None:
        ncvariables = list(self.ds.variables.keys())
        for key, values in dct.items():
            # Save origional for error message
            original = values[:]
            # Remove all values that are not in the netcdf
            for v in reversed(values):
                if v not in ncvariables:
                    values.remove(v)
            # One variable should remain
            if len(values) == 0:
                raise KeyError(
                    f'An attribute for "{key}" was not found in nc file. Expected "{"/".join(original)}"'
                )
            elif len(values) > 1:
                raise KeyError(
                    f'Multiple attributes for "{key}" were found in nc file. Got "{" and ".join(values)}"'
                )

            # Update the dictionary from a list to a single string
            dct[key] = values[0]



"""
TODO Implement the following structures
- Bridge
- Generalstructure
- Long culvert
- Gate
- Dambreak

Add comments for these structures. Maybe link them to descriptions of `Field`s?
"""

import logging
from enum import Enum
from pathlib import Path
from typing import List, Literal, Optional, Set, Union

from pydantic import Field
from pydantic.class_validators import root_validator

from hydrolib.core.io.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.io.ini.util import get_split_string_on_delimiter_validator

logger = logging.getLogger(__name__)

# TODO: handle comment blocks
# TODO: handle duplicate keys
class Structure(INIBasedModel):
    # TODO: would we want to load this from something externally and generate these automatically
    class Comments(INIBasedModel.Comments):
        id: Optional[str] = "Unique structure id (max. 256 characters)."
        name: Optional[str] = "Given name in the user interface."
        branchid: Optional[str] = Field(
            "Branch on which the structure is located.", alias="branchId"
        )
        chainage: Optional[str] = "Chainage on the branch (m)."

        n_coordinates: Optional[str] = Field(
            "Number of values in xCoordinates and yCoordinates", alias="numCoordinates"
        )
        x_coordinates: Optional[str] = Field(
            "x-coordinates of the location of the structure. (number of values = numCoordinates)",
            alias="xCoordinates",
        )
        y_coordinates: Optional[str] = Field(
            "y-coordinates of the location of the structure. (number of values = numCoordinates)",
            alias="yCoordinates",
        )

    comments: Comments = Comments()

    _header: Literal["Structure"] = "Structure"

    id: str = Field("id", max_length=256)
    name: str = Field("id")
    structure_type: str = Field(alias="type")

    branchid: Optional[str] = Field(None, alias="branchId")
    chainage: Optional[float] = None

    n_coordinates: Optional[int] = Field(None, alias="numCoordinates")
    x_coordinates: Optional[List[float]] = Field(None, alias="xCoordinates")
    y_coordinates: Optional[List[float]] = Field(None, alias="yCoordinates")

    @root_validator
    def check_location(cls, values):
        assert (
            "n_coordinates" in values
            and "x_coordinates" in values
            and "y_coordinates" in values
        ) or (
            "branchid" in values and "chainage" in values
        ), "Specify location either by setting `branchid` and `chainage` or `*_coordinates` fields."
        return values

    @classmethod
    def validate(cls, v):
        """Try to iniatialize subclass based on function field."""
        # should be replaced by discriminated unions once merged
        # https://github.com/samuelcolvin/pydantic/pull/2336
        if isinstance(v, dict):
            for c in cls.__subclasses__():
                if (
                    c.__fields__.get("structure_type").default
                    == v.get("type", "").lower()
                ):
                    v = c(**v)
                    break
            else:
                logger.warning(f"Couldn't derive specific type of {cls.__name__}")
        return super().validate(v)

    def _exclude_fields(self) -> Set:
        # exclude the unset props like coordinates or branches
        if self.branchid is not None:
            exclude_set = {"n_coordinates", "x_coordinates", "y_coordinates"}
        else:
            exclude_set = {"branchid", "chainage"}
        exclude_set = super()._exclude_fields().union(exclude_set)
        return exclude_set


class FlowDirection(str, Enum):
    none = "none"
    positive = "positive"
    negative = "negative"
    both = "both"


class Weir(Structure):
    class Comments(Structure.Comments):
        structure_type: Optional[str] = Field(
            "Structure type; must read weir", alias="type"
        )
        allowedflowdir: Optional[str] = Field(
            "Possible values: both, positive, negative, none.", alias="allowedFlowdir"
        )

        crestlevel: Optional[str] = Field(
            "Crest level of weir (m AD).", alias="crestLevel"
        )
        crestwidth: Optional[str] = Field("Width of the weir (m).", alias="crestWidth")
        corrcoeff: Optional[str] = Field(
            "Correction coefficient (-).", alias="corrCoeff"
        )
        usevelocityheight: Optional[str] = Field(
            "Flag indicating whether the velocity height is to be calculated or not.",
            alias="useVelocityHeight",
        )

    comments: Comments = Comments()

    structure_type: Literal["weir"] = Field("weir", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowdir")

    crestlevel: Union[float, Path] = Field(alias="crestLevel")
    crestwidth: Optional[float] = Field(None, alias="crestWidth")
    corrcoeff: float = Field(1.0, alias="corrCoeff")
    usevelocityheight: bool = Field(True, alias="useVelocityHeight")


class UniversalWeir(Structure):
    class Comments(Structure.Comments):
        structure_type: Optional[str] = Field(
            "Structure type; must read universalWeir", alias="type"
        )
        allowedflowdir: Optional[str] = Field(
            "Possible values: both, positive, negative, none.", alias="allowedFlowdir"
        )

        numlevels: Optional[str] = Field("Number of yz-Values.", alias="numLevels")
        yvalues: Optional[str] = Field(
            "y-values of the cross section (m). (number of values = numLevels)",
            alias="yValues",
        )
        zvalues: Optional[str] = Field(
            "z-values of the cross section (m). (number of values = numLevels)",
            alias="zValues",
        )
        crestlevel: Optional[str] = Field(
            "Crest level of weir (m AD).", alias="crestLevel"
        )
        dischargecoeff: Optional[str] = Field(
            "Discharge coefficient c_e (-).", alias="dischargeCoeff"
        )

    comments: Comments = Comments()

    structure_type: Literal["universalWeir"] = Field("universalWeir", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowdir")

    numlevels: int = Field(alias="numLevels")
    yvalues: List[float] = Field(alias="yValues")
    zvalues: List[float] = Field(alias="zValues")
    crestlevel: float = Field(alias="crestLevel")
    dischargecoeff: float = Field(alias="dischargeCoeff")

    _split_to_list = get_split_string_on_delimiter_validator("yvalues", "zvalues")


class CulvertSubType(str, Enum):
    culvert = "culvert"
    invertedSiphon = "invertedSiphon"


class Culvert(Structure):

    structure_type: Literal["culvert"] = Field("culvert", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowdir")

    leftlevel: float = Field(alias="leftLevel")
    rightlevel: float = Field(alias="rightLevel")
    csdefid: str = Field(alias="csDefId")
    length: float = Field(alias="length")
    inletlosscoeff: float = Field(alias="inletlossCoeff")
    outletlosscoeff: float = Field(alias="outletLossCoeff")
    valveonoff: bool = Field(alias="valveOnOff")
    valveopeningheight: Union[float, Path] = Field(alias="valveOpeningHeight")
    numlosscoeff: int = Field(alias="numLossCoeff")
    relopening: List[float] = Field(alias="relOpening")
    losscoeff: List[float] = Field(alias="lossCoeff")
    bedfrictiontype: str = Field(alias="bedFrictionType")
    bedfriction: float = Field(alias="bedFriction")
    subtype: CulvertSubType = Field(alias="subType")
    bendlosscoeff: float = Field(alias="bendLossCoeff")

    _split_to_list = get_split_string_on_delimiter_validator("relopening", "losscoeff")


class Pump(Structure):

    structure_type: Literal["pump"] = Field("pump", alias="type")

    orientation: str
    controlside: str = Field(alias="controlSide")  # TODO Enum
    numstages: int = Field(0, alias="numStages")
    capacity: Union[float, Path]

    startlevelsuctionside: List[float] = Field(alias="startLevelSuctionSide")
    stoplevelsuctionside: List[float] = Field(alias="stopLevelSuctionSide")
    startleveldeliveryside: List[float] = Field(alias="startLevelDeliverySide")
    stopleveldeliveryside: List[float] = Field(alias="stopLevelDeliverySide")
    numreductionlevels: int = Field(0, alias="numReductionLevels")
    head: List[float]
    reductionfactor: List[float] = Field(alias="reductionFactor")

    _split_to_list = get_split_string_on_delimiter_validator(
        "startlevelsuctionside",
        "stoplevelsuctionside",
        "startleveldeliveryside",
        "stopleveldeliveryside",
        "head",
        "reductionfactor",
    )


class Compound(Structure):

    structure_type: Literal["compound"] = Field("compound", alias="type")
    numstructures: int = Field(alias="numStructures")
    structureids: List[str] = Field(alias="structureIds")

    _split_to_list = get_split_string_on_delimiter_validator(
        "structureids", delimiter=";"
    )


class Orifice(Structure):

    structure_type: Literal["orifice"] = Field("orifice", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowdir")

    crestlevel: Union[float, Path] = Field(alias="crestLevel")
    crestwidth: Optional[float] = Field(None, alias="crestWidth")
    gateloweredgelevel: Union[float, Path] = Field(alias="gateLowerEdgeLevel")
    corrcoeff: float = Field(1.0, alias="corrCoeff")
    usevelocityheight: bool = Field(True, alias="useVelocityHeight")

    # TODO Use a validator here to check the optionals related to the bool field
    uselimitflowpos: bool = Field(False, alias="useLimitFlowPos")
    limitflowpos: Optional[float] = Field(alias="limitFlowPos")

    uselimitflowneg: bool = Field(False, alias="useLimitflowNeg")
    limitflowneg: Optional[float] = Field(alias="limitFlowneg")


class StuctureGeneral(INIGeneral):
    fileVersion: str = "3.00"
    fileType: Literal["structure"]


class StructureModel(INIModel):
    general: StuctureGeneral
    structure: List[Structure]

    @classmethod
    def _ext(cls) -> str:
        return ".ini"

    @classmethod
    def _filename(cls) -> str:
        return "structures"



"""
Here we define our Pydantic `BaseModel` with custom settings,
as well as a `FileModel` that inherits from a `BaseModel` but
also represents a file on disk.

"""
import logging
from abc import ABC, abstractclassmethod
from contextvars import ContextVar
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Type
from warnings import warn

from pydantic import BaseModel as PydanticBaseModel

from hydrolib.core.io.base import DummmyParser, DummySerializer
from hydrolib.core.utils import to_key

logger = logging.getLogger(__name__)

# We use ContextVars to keep a reference to the folder
# we're currently parsing files in. In the future
# we could move to https://github.com/samuelcolvin/pydantic/issues/1549
context_dir: ContextVar[Path] = ContextVar("folder")


def _reset_context_dir(token):
    # Once the model has been completely initialized
    # reset the context
    if token:
        logger.info("Reset context.")
        context_dir.reset(token)


class BaseModel(PydanticBaseModel):
    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True
        use_enum_values = True
        extra = "forbid"  # will throw errors so we can fix our models
        allow_population_by_field_name = True
        alias_generator = to_key

    def is_file_link(self) -> bool:
        """Generic attribute for models backed by a file."""
        return False

    def is_intermediate_link(self) -> bool:
        """Generic attribute for models that have children fields that could contain files."""
        return self.is_file_link()

    def show_tree(self, indent=0):
        """Recursive print function for showing a tree of a model."""
        angle = "∟" if indent > 0 else ""

        # Only print if we're backed by a file
        if self.is_file_link():
            print(" " * indent * 2, angle, self)

        # Otherwise we recurse through the fields of a model
        for _, value in self:
            # Handle lists of items
            if not isinstance(value, list):
                value = [value]
            for v in value:
                if hasattr(v, "is_intermediate_link") and v.is_intermediate_link():
                    # If the field is only an intermediate, print the name only
                    if not v.is_file_link():
                        print(" " * (indent * 2 + 2), angle, v.__class__.__name__)
                    v.show_tree(indent + 1)

    def _apply_recurse(self, f, *args, **kwargs):
        # TODO Could we use this function for `show_tree`?
        for _, value in self:
            # Handle lists of items
            if not isinstance(value, list):
                value = [value]
            for v in value:
                if hasattr(v, "is_intermediate_link") and v.is_intermediate_link():
                    if not v.is_file_link():
                        v._apply_recurse(f, *args, **kwargs)
                    else:
                        getattr(v, f)(*args, **kwargs)

        # Run self as last, so we can make use of the nested updates
        if self.is_file_link():
            getattr(self, f)(*args, **kwargs)


class FileModel(BaseModel, ABC):
    """Base class to represent models with a file representation.

    It therefore always has a `filepath` and if it is given on
    initilization, it will parse that file.

    This class extends the `validate` option of Pydantic,
    so when when a Path is given to a field with type `FileModel`,
    it doesn't error, but actually initializes the `FileModel`.
    """

    filepath: Optional[Path] = None

    def __init__(self, filepath: Optional[Path] = None, *args, **kwargs):
        """Initialize a model.

        The model is empty (with defaults) if no `filepath` is given,
        otherwise the file at `filepath` will be parsed."""
        # Parse the file if path is given
        context_dir_reset_token = None
        if filepath:
            filepath = Path(filepath)  # so we also accept strings

            # If a context is set, use it
            if (folder := context_dir.get(None)) and not filepath.is_absolute():
                logger.info(f"Used context to get {folder} for {filepath}")
                filepath = folder / filepath
            # Otherwise we're the root filepath
            # and should set the context
            else:
                logger.info(f"Set context to {filepath.parent}")
                context_dir_reset_token = context_dir.set(filepath.parent)

            data = self._load(filepath)
            data["filepath"] = filepath
            kwargs.update(data)
        super().__init__(*args, **kwargs)

        _reset_context_dir(context_dir_reset_token)

    def is_file_link(self) -> bool:
        return True

    @classmethod
    def validate(cls: Type["FileModel"], value: Any):
        # Enable initialization with a Path.
        if isinstance(value, (Path, str)):
            # Pydantic Model init requires a dict
            value = {"filepath": value}
        return super().validate(value)

    def _load(self, filepath: Path) -> Dict:
        # TODO Make this lazy in some cases
        # so it doesn't become slow
        if filepath.is_file():
            return self._parse(filepath)
        else:
            warn(f"File: `{filepath}` not found, skipped parsing.")
            return {}

    def save(self, folder: Optional[Path] = None) -> Path:
        """Save model and child models to their set filepaths.

        If a folder is given, for models with an unset filepath,
        we generate one based on the given `folder` and a default name.
        Otherwise we override the folder part of already set filepaths.
        This can thus be used to copy complete models.

        Args:
            folder: path to the folder where this FileModel will be stored
        """
        if not self.filepath and not folder:
            raise ValueError(
                "Either set the `filepath` on the model or pass a `folder` when saving."
            )

        if not folder:
            folder = self.filepath.absolute().parent

        self._apply_recurse("_save", folder)
        return self.filepath.absolute()

    def _save(self, folder):
        filename = Path(self.filepath.name) if self.filepath else self._generate_name()
        self.filepath = folder / filename

        self._serialize(self.dict())

    def _serialize(self, data: dict) -> None:
        self._get_serializer()(self.filepath, data)

    @classmethod
    def _parse(cls, path: Path) -> Dict:
        return cls._get_parser()(path)

    @classmethod
    def _generate_name(cls) -> Path:
        name, ext = cls._filename(), cls._ext()
        return Path(f"{name}{ext}")

    @abstractclassmethod
    def _filename(cls):
        return "test"

    @abstractclassmethod
    def _ext(cls):
        return ".test"

    @abstractclassmethod
    def _get_serializer(cls) -> Callable:
        return DummySerializer.serialize

    @abstractclassmethod
    def _get_parser(cls) -> Callable:
        return DummmyParser.parse

    def __str__(self) -> str:
        return str(self.filepath if self.filepath else "")



from abc import ABC
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, List, Literal

from pydantic.fields import Field

from hydrolib.core.io.ini.models import DataBlockINIBasedModel, INIGeneral, INIModel
from hydrolib.core.io.ini.parser import Parser, ParserConfig
from hydrolib.core.io.ini.serializer import SerializerConfig, write_ini
from hydrolib.core.io.ini.util import make_list_validator


class VerticalInterpolation(str, Enum):
    linear = "linear"
    log = "log"
    block = "block"


class VerticalPositionType(str, Enum):
    percentage_bed = "percBed"
    z_bed = "ZBed"


class TimeInterpolation(str, Enum):
    linear = "linear"
    block_from = "blockFrom"
    block_to = "blockTo"


class ForcingBase(DataBlockINIBasedModel):

    _header: Literal["forcing"] = "forcing"
    name: str
    function: str
    quantity: List[str]
    unit: List[str]

    _make_lists = make_list_validator("quantity", "unit")

    @classmethod
    def _supports_comments(cls):
        return False

    @classmethod
    def _duplicate_keys_as_list(cls):
        return True

    @classmethod
    def validate(cls, v):
        """Try to iniatialize subclass based on function field."""
        # should be replaced by discriminated unions once merged
        # https://github.com/samuelcolvin/pydantic/pull/2336
        if isinstance(v, dict):
            for c in cls.__subclasses__():
                if (
                    c.__fields__.get("function").default
                    == v.get("function", "").lower()
                ):
                    v = c(**v)
                    break
        return v


class TimeSeries(ForcingBase):
    function: Literal["timeseries"] = "timeseries"
    timeinterpolation: TimeInterpolation = Field(alias="timeInterpolation")
    offset: float = 0.0
    factor: float = 1.0


class Harmonic(ForcingBase):
    function: Literal["harmonic"] = "harmonic"
    factor: float = 1.0


class Astronomic(ForcingBase):
    function: Literal["astronomic"] = "astronomic"
    factor: float = 1.0


class HarmonicCorrection(ForcingBase):
    function: Literal["harmoniccorrection"] = "harmoniccorrection"


class AstronomicCorrection(ForcingBase):
    function: Literal["astronomiccorrection"] = "astronomiccorrection"


class T3D(ForcingBase):
    function: Literal["t3d"] = "t3d"

    offset: float = 0.0
    factor: float = 1.0

    vertical_positions: List[float] = Field(alias="verticalPositions")
    vertical_interpolation: VerticalInterpolation = Field(alias="verticalInterpolation")
    vertical_position_type: VerticalPositionType = Field(alias="verticalPositionType")


class QHTable(ForcingBase):
    function: Literal["qhtable"] = "qhtable"


class Constant(ForcingBase):
    function: Literal["constant"] = "constant"

    offset: float = 0.0
    factor: float = 1.0


class ForcingGeneral(INIGeneral):
    fileVersion: str = "1.01"
    fileType: Literal["boundConds"] = "boundConds"


class ForcingModel(INIModel):
    general: ForcingGeneral
    forcing: List[ForcingBase]

    @classmethod
    def _ext(cls) -> str:
        return ".bc"

    @classmethod
    def _filename(cls) -> str:
        return "forcing"

    @classmethod
    def _get_parser(cls) -> Callable:
        return cls.parse

    @classmethod
    def parse(cls, filepath: Path):
        # It's odd to have to disable parsing something as comments
        # but also need to pass it to the *flattener*.
        # This method now only supports per model settings, not per section.
        parser = Parser(ParserConfig(parse_datablocks=True, parse_comments=False))

        with filepath.open() as f:
            for line in f:
                parser.feed_line(line)

        return parser.finalize().flatten(True, False)

    def _serialize(self, _: dict) -> None:
        # We skip the passed dict for a better one.
        config = SerializerConfig(section_indent=0, property_indent=4)
        write_ini(self.filepath, self._to_document(), config=config)



from hydrolib.core.io.net.models import NetworkModel
from pathlib import Path
from typing import Callable, List, Literal, Optional, Union
from pydantic import Field
from hydrolib.core.basemodel import FileModel
from hydrolib.core.io.base import DummySerializer
from hydrolib.core.io.bc.models import ForcingModel
from hydrolib.core.io.ini.models import (
    CrossDefModel,
    CrossLocModel,
    FrictionModel,
    INIBasedModel,
    INIGeneral,
    INIModel,
)
from hydrolib.core.io.ini.parser import Parser
from hydrolib.core.io.ini.serializer import SerializerConfig, write_ini
from hydrolib.core.io.ini.util import (
    get_split_string_on_delimiter_validator,
    make_list_validator,
)
from hydrolib.core.io.polyfile.models import PolyFile
from hydrolib.core.io.structure.models import StructureModel
from hydrolib.core.io.xyz.models import XYZModel


class General(INIBasedModel):
    _header: Literal["general"] = "general"
    program: str = Field("D-Flow FM", alias="Program")
    version: str = Field("1.2.94.66079M", alias="Version")
    filetype: Literal["modelDef"] = Field("modelDef", alias="fileType")
    fileversion: str = Field("1.09", alias="fileVersion")
    autostart: bool = Field(False, alias="AutoStart")
    pathsrelativetoparent: bool = Field(False, alias="PathsRelativeToParent")


class Numerics(INIBasedModel):
    _header: Literal["numerics"] = "numerics"
    cflmax: float = Field(0.7, alias="CFLMax")
    advectype: int = Field(33, alias="AdvecType")
    timesteptype: int = Field(2, alias="TimeStepType")
    limtyphu: int = Field(0, alias="Limtyphu")
    limtypmom: int = Field(4, alias="Limtypmom")
    limtypsa: int = Field(4, alias="Limtypsa")
    icgsolver: int = Field(4, alias="Icgsolver")
    maxdegree: int = Field(6, alias="Maxdegree")
    fixedweirscheme: int = Field(9, alias="FixedWeirScheme")
    fixedweircontraction: float = Field(1.0, alias="FixedWeirContraction")
    izbndpos: int = Field(0, alias="Izbndpos")
    tlfsmo: float = Field(0.0, alias="Tlfsmo")
    slopedrop2d: float = Field(0.0, alias="Slopedrop2D")
    drop1d: bool = Field(False, alias="drop1D")
    chkadvd: float = Field(0.1, alias="Chkadvd")
    teta0: float = Field(0.55, alias="Teta0")
    qhrelax: float = Field(0.01, alias="Qhrelax")
    cstbnd: bool = Field(False, alias="cstbnd")
    maxitverticalforestersal: int = Field(0, alias="Maxitverticalforestersal")
    maxitverticalforestertem: int = Field(0, alias="Maxitverticalforestertem")
    turbulencemodel: int = Field(3, alias="Turbulencemodel")
    turbulenceadvection: int = Field(3, alias="Turbulenceadvection")
    anticreep: bool = Field(False, alias="AntiCreep")
    maxwaterleveldiff: float = Field(0.0, alias="Maxwaterleveldiff")
    maxvelocitydiff: float = Field(0.0, alias="Maxvelocitydiff")
    epshu: float = Field(0.0001, alias="Epshu")


class VolumeTables(INIBasedModel):
    _header: Literal["volumeTables"] = "volumeTables"
    usevolumetables: bool = Field(False, alias="useVolumeTables")
    increment: float = Field(0.2, alias="increment")
    usevolumetablefile: bool = Field(False, alias="useVolumeTableFile")


class Physics(INIBasedModel):
    _header: Literal["physics"] = "physics"
    uniffrictcoef: float = Field(0.023, alias="UnifFrictCoef")
    uniffricttype: int = Field(1, alias="UnifFrictType")
    uniffrictcoef1d: float = Field(0.023, alias="UnifFrictCoef1D")
    uniffrictcoeflin: float = Field(0.0, alias="UnifFrictCoefLin")
    umodlin: int = Field(0, alias="Umodlin")
    vicouv: float = Field(0.1, alias="Vicouv")
    dicouv: float = Field(0.1, alias="Dicouv")
    vicoww: float = Field(5e-05, alias="Vicoww")
    dicoww: float = Field(5e-05, alias="Dicoww")
    vicwminb: float = Field(0.0, alias="Vicwminb")
    xlozmidov: float = Field(0.0, alias="Xlozmidov")
    smagorinsky: float = Field(0.2, alias="Smagorinsky")
    elder: float = Field(0.0, alias="Elder")
    irov: int = Field(0, alias="irov")
    wall_ks: float = Field(0.0, alias="wall_ks")
    rhomean: float = Field(1000, alias="Rhomean")
    idensform: int = Field(2, alias="Idensform")
    ag: float = Field(9.81, alias="Ag")
    tidalforcing: bool = Field(False, alias="TidalForcing")
    doodsonstart: float = Field(55.565, alias="Doodsonstart")
    doodsonstop: float = Field(375.575, alias="Doodsonstop")
    doodsoneps: float = Field(0.0, alias="Doodsoneps")
    villemontecd1: float = Field(1.0, alias="VillemonteCD1")
    villemontecd2: float = Field(10.0, alias="VillemonteCD2")
    salinity: bool = Field(False, alias="Salinity")
    initialsalinity: float = Field(0.0, alias="InitialSalinity")
    sal0abovezlev: float = Field(-999.0, alias="Sal0abovezlev")
    deltasalinity: float = Field(-999.0, alias="DeltaSalinity")
    backgroundsalinity: float = Field(30.0, alias="Backgroundsalinity")
    temperature: int = Field(0, alias="Temperature")
    initialtemperature: float = Field(6.0, alias="InitialTemperature")
    backgroundwatertemperature: float = Field(6.0, alias="BackgroundwaterTemperature")
    secchidepth: float = Field(2.0, alias="Secchidepth")
    stanton: float = Field(0.0013, alias="Stanton")
    dalton: float = Field(0.0013, alias="Dalton")
    secondaryflow: bool = Field(False, alias="SecondaryFlow")
    betaspiral: int = Field(0, alias="BetaSpiral")


class Wind(INIBasedModel):
    _header: Literal["wind"] = "wind"
    icdtyp: int = Field(2, alias="ICdtyp")
    cdbreakpoints: List[float] = Field([0.00063, 0.00723], alias="Cdbreakpoints")
    windspeedbreakpoints: List[float] = Field(
        [0.0, 100.0], alias="Windspeedbreakpoints"
    )
    rhoair: float = Field(1.205, alias="Rhoair")
    relativewind: bool = Field(False, alias="Relativewind")
    windpartialdry: bool = Field(True, alias="Windpartialdry")
    pavbnd: float = Field(0.0, alias="PavBnd")
    pavini: float = Field(0.0, alias="PavIni")

    @classmethod
    def list_delimiter(cls) -> str:
        return " "

    _split_to_list = get_split_string_on_delimiter_validator(
        "cdbreakpoints",
        "windspeedbreakpoints",
    )


class Waves(INIBasedModel):
    _header: Literal["waves"] = "waves"
    wavemodelnr: int = Field(3, alias="Wavemodelnr")
    rouwav: str = Field("FR84", alias="Rouwav")
    gammax: float = Field(0.5, alias="Gammax")


class Time(INIBasedModel):
    _header: Literal["time"] = "time"
    refdate: int = Field(20200101, alias="RefDate")  # TODO Convert to datetime
    tzone: float = Field(0.0, alias="Tzone")
    tunit: str = Field("S", alias="Tunit")  # DHMS
    dtuser: float = Field(300.0, alias="DtUser")
    dtnodal: float = Field(21600.0, alias="DtNodal")
    dtmax: float = Field(30.0, alias="DtMax")
    dtinit: float = Field(1.0, alias="DtInit")
    tstart: float = Field(0.0, alias="TStart")
    tstop: float = Field(86400.0, alias="TStop")
    updateroughnessinterval: float = Field(86400.0, alias="UpdateRoughnessInterval")


class Restart(INIBasedModel):
    _header: Literal["restart"] = "restart"
    restartfile: Optional[Path] = Field(None, alias="RestartFile")
    restartdatetime: Optional[str] = Field(None, alias="RestartDateTime")


class Boundary(INIBasedModel):
    _header: Literal["Boundary"] = "Boundary"
    quantity: str
    nodeId: str
    forcingfile: Optional[ForcingModel] = "FlowFM_boundaryconditions1d.bc"
    bndWidth1D: float

    def is_intermediate_link(self) -> bool:
        return True


class Lateral(INIBasedModel):
    _header: Literal["Lateral"] = "Lateral"
    id: str
    name: str
    nodeId: str
    discharge: str


class ExtGeneral(INIGeneral):
    _header: Literal["General"] = "General"
    fileVersion: str = "2.01"
    fileType: Literal["extForce"] = "extForce"


class ExtModel(INIModel):
    general: ExtGeneral
    boundary: List[Boundary]
    lateral: List[Lateral]

    _split_to_list = make_list_validator("boundary", "lateral")

    @classmethod
    def _ext(cls) -> str:
        return ".ext"

    @classmethod
    def _filename(cls) -> str:
        return "bnd"

    def _serialize(self, _: dict) -> None:
        # We skip the passed dict for a better one.
        config = SerializerConfig(section_indent=0, property_indent=0)
        write_ini(self.filepath, self._to_document(), config=config)


class ExternalForcing(INIBasedModel):
    _header: Literal["external forcing"] = "external forcing"
    extforcefile: Optional[Path] = Field(None, alias="ExtForceFile")
    extforcefilenew: Optional[ExtModel] = Field(None, alias="ExtForceFileNew")
    rainfall: Optional[bool] = Field(None, alias="Rainfall")
    qext: Optional[bool] = Field(None, alias="QExt")
    evaporation: Optional[bool] = Field(None, alias="Evaporation")
    windext: Optional[int] = Field(None, alias="WindExt")

    def is_intermediate_link(self) -> bool:
        return True


class Hydrology(INIBasedModel):
    _header: Literal["hydrology"] = "hydrology"
    interceptionmodel: bool = Field(False, alias="InterceptionModel")


class Trachytopes(INIBasedModel):
    _header: Literal["trachytopes"] = "trachytopes"
    trtrou: str = Field("N", alias="TrtRou")  # TODO bool
    trtdef: Optional[int] = Field(None, alias="TrtDef")  # no doc?
    trtl: Optional[int] = Field(None, alias="TrtL")  # no doc?
    dttrt: float = Field(60.0, alias="DtTrt")


class Output(INIBasedModel):
    _header: Literal["output"] = "output"
    wrishp_crs: bool = Field(False, alias="Wrishp_crs")
    wrishp_weir: bool = Field(False, alias="Wrishp_weir")
    wrishp_gate: bool = Field(False, alias="Wrishp_gate")
    wrishp_fxw: bool = Field(False, alias="Wrishp_fxw")
    wrishp_thd: bool = Field(False, alias="Wrishp_thd")
    wrishp_obs: bool = Field(False, alias="Wrishp_obs")
    wrishp_emb: bool = Field(False, alias="Wrishp_emb")
    wrishp_dryarea: bool = Field(False, alias="Wrishp_dryarea")
    wrishp_enc: bool = Field(False, alias="Wrishp_enc")
    wrishp_src: bool = Field(False, alias="Wrishp_src")
    wrishp_pump: bool = Field(False, alias="Wrishp_pump")
    outputdir: Optional[Path] = Field(None, alias="OutputDir")
    waqoutputdir: Optional[Path] = Field(None, alias="WAQOutputDir")
    flowgeomfile: Optional[Path] = Field(None, alias="FlowGeomFile")
    obsfile: Optional[List[Path]] = Field(None, alias="ObsFile")
    crsfile: Optional[List[Path]] = Field(None, alias="CrsFile")
    hisfile: Optional[Path] = Field(None, alias="HisFile")
    hisinterval: float = Field(300, alias="HisInterval")
    xlsinterval: float = Field(0.0, alias="XLSInterval")
    mapfile: Optional[Path] = Field(None, alias="MapFile")
    mapinterval: float = Field(1200.0, alias="MapInterval")
    rstinterval: float = Field(0.0, alias="RstInterval")
    mapformat: int = Field(4, alias="MapFormat")
    ncformat: int = Field(3, alias="NcFormat")
    ncnounlimited: bool = Field(False, alias="NcNoUnlimited")
    ncnoforcedflush: bool = Field(False, alias="NcNoForcedFlush")
    ncwritelatlon: bool = Field(False, alias="NcWriteLatLon")

    # his file
    wrihis_balance: bool = Field(True, alias="Wrihis_balance")
    wrihis_sourcesink: bool = Field(True, alias="Wrihis_sourcesink")
    wrihis_structure_gen: bool = Field(True, alias="Wrihis_structure_gen")
    wrihis_structure_dam: bool = Field(True, alias="Wrihis_structure_dam")
    wrihis_structure_pump: bool = Field(True, alias="Wrihis_structure_pump")
    wrihis_structure_gate: bool = Field(True, alias="Wrihis_structure_gate")
    wrihis_structure_weir: bool = Field(True, alias="Wrihis_structure_weir")
    wrihis_structure_orifice: bool = Field(True, alias="Wrihis_structure_orifice")
    wrihis_structure_bridge: bool = Field(True, alias="Wrihis_structure_bridge")
    wrihis_structure_culvert: bool = Field(True, alias="Wrihis_structure_culvert")
    wrihis_structure_longculvert: bool = Field(
        True, alias="Wrihis_structure_longculvert"
    )
    wrihis_structure_dambreak: bool = Field(True, alias="Wrihis_structure_damBreak")
    wrihis_structure_uniweir: bool = Field(True, alias="Wrihis_structure_uniWeir")
    wrihis_structure_compound: bool = Field(True, alias="Wrihis_structure_compound")
    wrihis_lateral: bool = Field(True, alias="Wrihis_lateral")
    wrihis_velocity: bool = Field(False, alias="Wrihis_velocity")
    wrihis_discharge: bool = Field(False, alias="Wrihis_discharge")

    # Map file
    wrimap_waterlevel_s0: bool = Field(True, alias="Wrimap_waterlevel_s0")
    wrimap_waterlevel_s1: bool = Field(True, alias="Wrimap_waterlevel_s1")
    wrimap_evaporation: bool = Field(True, alias="Wrimap_evaporation")
    wrimap_velocity_component_u0: bool = Field(
        True, alias="Wrimap_velocity_component_u0"
    )
    wrimap_velocity_component_u1: bool = Field(
        True, alias="Wrimap_velocity_component_u1"
    )
    wrimap_velocity_vector: bool = Field(True, alias="Wrimap_velocity_vector")
    wrimap_upward_velocity_component: bool = Field(
        False, alias="Wrimap_upward_velocity_component"
    )
    wrimap_density_rho: bool = Field(True, alias="Wrimap_density_rho")
    wrimap_horizontal_viscosity_viu: bool = Field(
        True, alias="Wrimap_horizontal_viscosity_viu"
    )
    wrimap_horizontal_diffusivity_diu: bool = Field(
        True, alias="Wrimap_horizontal_diffusivity_diu"
    )
    wrimap_flow_flux_q1: bool = Field(True, alias="Wrimap_flow_flux_q1")
    wrimap_spiral_flow: bool = Field(True, alias="Wrimap_spiral_flow")
    wrimap_numlimdt: bool = Field(True, alias="Wrimap_numlimdt")
    wrimap_taucurrent: bool = Field(True, alias="Wrimap_taucurrent")
    wrimap_chezy: bool = Field(True, alias="Wrimap_chezy")
    wrimap_turbulence: bool = Field(True, alias="Wrimap_turbulence")
    wrimap_rain: bool = Field(False, alias="Wrimap_rain")
    wrimap_wind: bool = Field(True, alias="Wrimap_wind")
    wrimap_heat_fluxes: bool = Field(False, alias="Wrimap_heat_fluxes")
    wrimap_wet_waterdepth_threshold: float = Field(
        2e-5, alias="Wrimap_wet_waterdepth_threshold"
    )
    wrimap_time_water_on_ground: bool = Field(
        False, alias="Wrimap_time_water_on_ground"
    )
    wrimap_freeboard: bool = Field(False, alias="Wrimap_freeboard")
    wrimap_waterdepth_on_ground: bool = Field(
        False, alias="Wrimap_waterdepth_on_ground"
    )
    wrimap_volume_on_ground: bool = Field(False, alias="Wrimap_volume_on_ground")
    wrimap_total_net_inflow_1d2d: bool = Field(
        False, alias="Wrimap_total_net_inflow_1d2d"
    )
    wrimap_total_net_inflow_lateral: bool = Field(
        False, alias="Wrimap_total_net_inflow_lateral"
    )
    wrimap_water_level_gradient: bool = Field(
        False, alias="Wrimap_water_level_gradient"
    )
    wrimap_flow_analysis: bool = Field(False, alias="Wrimap_flow_analysis")
    mapoutputtimevector: Optional[Path] = Field(None, alias="MapOutputTimeVector")
    fullgridoutput: bool = Field(False, alias="FullGridOutput")
    eulervelocities: bool = Field(False, alias="EulerVelocities")
    classmapfile: Optional[Path] = Field(None, alias="ClassMapFile")
    waterlevelclasses: List[float] = Field([0.0], alias="WaterlevelClasses")
    waterdepthclasses: List[float] = Field([0.0], alias="WaterdepthClasses")
    classmapinterval: float = Field(0.0, alias="ClassMapInterval")
    waqinterval: float = Field(0.0, alias="WaqInterval")
    statsinterval: float = Field(0.0, alias="StatsInterval")
    writebalancefile: bool = Field(False, alias="Writebalancefile")
    timingsinterval: float = Field(0.0, alias="TimingsInterval")
    richardsononoutput: bool = Field(True, alias="Richardsononoutput")

    _split_to_list = get_split_string_on_delimiter_validator(
        "waterlevelclasses",
        "waterdepthclasses",
        delimiter=";",
    )

    def is_intermediate_link(self) -> bool:
        # TODO set to True once we replace Paths with FileModels
        return False


class Geometry(INIBasedModel):

    _header: Literal["geometry"] = "geometry"
    netfile: Optional[NetworkModel] = Field(None, alias="NetFile")
    bathymetryfile: Optional[XYZModel] = Field(None, alias="BathymetryFile")
    drypointsfile: Optional[List[Union[XYZModel, PolyFile]]] = Field(
        None, alias="DryPointsFile"
    )  # TODO Fix, this will always try XYZ first, alias="]")
    structurefile: Optional[List[StructureModel]] = Field(None, alias="StructureFile")
    inifieldfile: Optional[Path] = Field(None, alias="IniFieldFile")
    waterlevinifile: Optional[Path] = Field(None, alias="WaterLevIniFile")
    landboundaryfile: Optional[List[Path]] = Field(None, alias="LandBoundaryFile")
    thindamfile: Optional[List[PolyFile]] = Field(None, alias="ThinDamFile")
    fixedweirfile: Optional[List[PolyFile]] = Field(None, alias="FixedWeirFile")
    pillarfile: Optional[List[PolyFile]] = Field(None, alias="PillarFile")
    usecaching: bool = Field(False, alias="UseCaching")
    vertplizfile: Optional[PolyFile] = Field(None, alias="VertplizFile")
    frictfile: Optional[List[FrictionModel]] = Field(None, alias="FrictFile")
    crossdeffile: Optional[CrossDefModel] = Field(None, alias="CrossDefFile")
    crosslocfile: Optional[CrossLocModel] = Field(None, alias="CrossLocFile")
    storagenodefile: Optional[Path] = Field(None, alias="StorageNodeFile")
    onedtwodlinkfile: Optional[Path] = Field(None, alias="oneDtwoDLinkFile")
    proflocfilefile: Optional[Path] = Field(None, alias="ProflocFileFile")
    profdeffile: Optional[Path] = Field(None, alias="ProfdefFile")
    profdefxyzfile: Optional[Path] = Field(None, alias="ProfdefxyzFile")
    manholefile: Optional[Path] = Field(None, alias="ManholeFile")
    partitionfile: Optional[PolyFile] = Field(None, alias="PartitionFile")
    uniformwidth1d: float = Field(2.0, alias="Uniformwidth1D")
    waterlevini: float = Field(0.0, alias="WaterLevIni")
    bedlevuni: float = Field(-5.0, alias="Bedlevuni")
    bedslope: float = Field(0.0, alias="Bedslope")
    bedlevtype: int = Field(3, alias="BedlevType")
    blmeanbelow: float = Field(-999.0, alias="Blmeanbelow")
    blminabove: float = Field(-999.0, alias="Blminabove")
    anglat: float = Field(0.0, alias="AngLat")
    anglon: float = Field(0.0, alias="AngLon")
    conveyance2d: int = Field(-1, alias="Conveyance2D")
    nonlin1d: int = Field(1, alias="Nonlin1D")
    nonlin2d: int = Field(0, alias="Nonlin2D")
    sillheightmin: float = Field(0.0, alias="Sillheightmin")
    makeorthocenters: bool = Field(False, alias="Makeorthocenters")
    dcenterinside: float = Field(1.0, alias="Dcenterinside")
    bamin: float = Field(1e-06, alias="Bamin")
    openboundarytolerance: float = Field(3.0, alias="OpenBoundaryTolerance")
    renumberflownodes: bool = Field(True, alias="RenumberFlowNodes")
    kmx: int = Field(0, alias="Kmx")
    layertype: int = Field(1, alias="Layertype")
    numtopsig: int = Field(0, alias="Numtopsig")
    sigmagrowthfactor: float = Field(1.0, alias="SigmaGrowthFactor")
    dxdoubleat1dendnodes: bool = Field(True, alias="dxDoubleAt1DEndNodes")
    changevelocityatstructures: bool = Field(False, alias="ChangeVelocityAtStructures")
    changestructuredimensions: bool = Field(True, alias="ChangeStructureDimensions")

    _split_to_list = get_split_string_on_delimiter_validator(
        "frictfile",
        "structurefile",
        "drypointsfile",
        "landboundaryfile",
        "thindamfile",
        "fixedweirfile",
        "pillarfile",
        delimiter=";",
    )

    def is_intermediate_link(self) -> bool:
        return True


class FMModel(INIModel):
    """FM Model representation."""

    general: General = General()
    geometry: Geometry = Geometry()
    volumetables: VolumeTables = VolumeTables()
    numerics: Numerics = Numerics()
    physics: Physics = Physics()
    wind: Wind = Wind()
    waves: Optional[Waves] = None
    time: Time = Time()
    restart: Restart = Restart()
    external_forcing: ExternalForcing = ExternalForcing()
    hydrology: Hydrology = Hydrology()
    trachytopes: Trachytopes = Trachytopes()
    output: Output = Output()

    @classmethod
    def _ext(cls) -> str:
        return ".mdu"

    @classmethod
    def _filename(cls) -> str:
        return "fm"



"""util.py provides additional utility methods related to handling ini files.
"""
from typing import Any

from pydantic.class_validators import validator


def get_split_string_on_delimiter_validator(*field_name: str, delimiter: str = None):
    """Get a validator to split strings passed to the specified field_name.

    Strings are split based on the provided delimeter. The validator splits
    strings into a list of substrings before any other validation takes
    place.

    Args:
        delimiter (str, optional):
            The delimiter to split the string with. Defaults to None.

    Returns:
        the validator which splits strings on the provided delimiter.
    """

    def split(v: Any):
        if isinstance(v, str):
            v = v.split(delimiter)
        return v

    return validator(*field_name, allow_reuse=True, pre=True)(split)


def make_list_validator(*field_name: str):
    """Get a validator make a list of object if a single object is passed."""

    def split(v: Any):
        if not isinstance(v, list):
            v = [v]
        return v

    return validator(*field_name, allow_reuse=True, pre=True)(split)



from __future__ import annotations

from pathlib import Path
from typing import List, Optional, Union, Dict

import meshkernel as mk
from meshkernel.py_structures import GeometryList
import netCDF4 as nc
import numpy as np
from pydantic import Field

from hydrolib.core import __version__
from hydrolib.core.basemodel import BaseModel, FileModel


def split_by(gl: mk.GeometryList, by: float) -> list:
    """Function to split mk.GeometryList by seperator."""
    x, y = gl.x_coordinates.copy(), gl.y_coordinates.copy()
    idx = np.where(x == by)[0]

    xparts = np.split(x, idx)
    yparts = np.split(y, idx)

    lists = [
        mk.GeometryList(xp[min(i, 1) :], yp[min(i, 1) :])
        for i, (xp, yp) in enumerate(zip(xparts, yparts))
    ]

    return lists


class Mesh2d(BaseModel):
    meshkernel: Optional[mk.MeshKernel] = Field(default_factory=mk.MeshKernel)

    mesh2d_node_x: np.ndarray = np.empty(0, dtype=np.double)
    mesh2d_node_y: np.ndarray = np.empty(0, dtype=np.double)
    mesh2d_node_z: np.ndarray = np.empty(0, dtype=np.double)

    mesh2d_edge_x: np.ndarray = np.empty(0, dtype=np.double)
    mesh2d_edge_y: np.ndarray = np.empty(0, dtype=np.double)
    mesh2d_edge_z: np.ndarray = np.empty(0, dtype=np.double)
    mesh2d_edge_nodes: np.ndarray = np.empty((0, 2), dtype=np.int32)

    mesh2d_face_x: np.ndarray = np.empty(0, dtype=np.double)
    mesh2d_face_y: np.ndarray = np.empty(0, dtype=np.double)
    mesh2d_face_z: np.ndarray = np.empty(0, dtype=np.double)
    mesh2d_face_nodes: np.ndarray = np.empty((0, 0), dtype=np.int32)

    def is_empty(self) -> bool:
        return self.mesh2d_node_x.size == 0

    def read_file(self, file_path: Path) -> None:
        from hydrolib.core.io.net.reader import UgridReader

        reader = UgridReader(file_path)
        reader.read_mesh2d(self)

    def _set_mesh2d(self):
        mesh2d = mk.Mesh2d(
            node_x=self.mesh2d_node_x,
            node_y=self.mesh2d_node_y,
            edge_nodes=self.mesh2d_edge_nodes.ravel(),
        )

        self.meshkernel.mesh2d_set(mesh2d)

    def get_mesh2d(self) -> mk.Mesh2d:
        """Return mesh2d from meshkernel"""
        return self.meshkernel.mesh2d_get()

    def create_rectilinear(self, extent: tuple, dx: float, dy: float) -> None:
        """Create a rectilinear mesh within a polygon. A rectangular grid is generated within the polygon bounds

        Args:
            extent (tuple): Bounding box of mesh (left, bottom, right, top)
            dx (float): Horizontal distance
            dy (float): Vertical distance

        TODO: Perhaps the polygon processing part should be part of Hydrolib (not core)!

        Raises:
            NotImplementedError: MultiPolygons
        """

        xmin, ymin, xmax, ymax = extent

        # Generate mesh
        mesh2d_input = mk.Mesh2dFactory.create_rectilinear_mesh(
            rows=int((ymax - ymin) / dy),
            columns=int((xmax - xmin) / dx),
            origin_x=xmin,
            origin_y=ymin,
            spacing_x=dx,
            spacing_y=dy,
        )

        # Process
        self._process(mesh2d_input)

    def _process(self, mesh2d_input):
        # Add input
        self.meshkernel.mesh2d_set(mesh2d_input)
        # Get output
        mesh2d_output = self.meshkernel.mesh2d_get()
        # Add to mesh2d variables
        self.mesh2d_node_x = mesh2d_output.node_x
        self.mesh2d_node_y = mesh2d_output.node_y

        self.mesh2d_edge_x = mesh2d_output.edge_x
        self.mesh2d_edge_y = mesh2d_output.edge_y
        self.mesh2d_edge_nodes = mesh2d_output.edge_nodes.reshape((-1, 2))

        self.mesh2d_face_x = mesh2d_output.face_x
        self.mesh2d_face_y = mesh2d_output.face_y
        npf = mesh2d_output.nodes_per_face
        self.mesh2d_face_nodes = np.full(
            (len(self.mesh2d_face_x), max(npf)), np.iinfo(np.int32).min
        )
        idx = (
            np.ones_like(self.mesh2d_face_nodes) * np.arange(max(npf))[None, :]
        ) < npf[:, None]
        self.mesh2d_face_nodes[idx] = mesh2d_output.face_nodes

    def clip(self, polygon: mk.GeometryList, deletemeshoption: int = 1):
        """Clip the 2D mesh by a polygon. Both outside the exterior and inside the interiors is clipped

        Args:
            polygon (GeometryList): Polygon stored as GeometryList
            deletemeshoption (int, optional): [description]. Defaults to 1.
        """

        # Add current mesh to Mesh2d instance
        self._set_mesh2d()

        # Delete outside polygon
        deletemeshoption = mk.DeleteMeshOption(deletemeshoption)
        parts = split_by(polygon, -998.0)

        # Check if parts are closed
        for part in parts:
            if not (part.x_coordinates[0], part.y_coordinates[0]) == (
                part.x_coordinates[-1],
                part.y_coordinates[-1],
            ):
                raise ValueError(
                    "First and last coordinate of each GeometryList part should match."
                )

        self.meshkernel.mesh2d_delete(parts[0], deletemeshoption, True)

        # Delete all holes
        for interior in parts[1:]:
            self.meshkernel.mesh2d_delete(interior, deletemeshoption, False)

        # Process
        self._process(self.meshkernel.mesh2d_get())

    def refine(self, polygon: mk.GeometryList, level: int):
        """Refine the mesh within a polygon, by a number of steps (level)

        Args:
            polygon (GeometryList): Polygon in which to refine
            level (int): Number of refinement steps
        """
        # Add current mesh to Mesh2d instance
        mesh2d_input = mk.Mesh2d(
            node_x=self.mesh2d_node_x,
            node_y=self.mesh2d_node_y,
            edge_nodes=self.mesh2d_edge_nodes.ravel(),
        )
        self.meshkernel.mesh2d_set(mesh2d_input)

        # Check if parts are closed
        if not (polygon.x_coordinates[0], polygon.y_coordinates[0]) == (
            polygon.x_coordinates[-1],
            polygon.y_coordinates[-1],
        ):
            raise ValueError(
                "First and last coordinate of each GeometryList part should match."
            )

        parameters = mk.MeshRefinementParameters(
            refine_intersected=True,
            use_mass_center_when_refining=False,
            min_face_size=10.0,  # Does nothing?
            refinement_type=1,
            connect_hanging_nodes=True,
            account_for_samples_outside_face=False,
            max_refinement_iterations=level,
        )
        self.meshkernel.mesh2d_refine_based_on_polygon(polygon, parameters)

        # Process
        self._process(self.meshkernel.mesh2d_get())


class Branch:

    # geometry: np.empty(0, dtype=np.double)
    # _x_coordinates: np.empty(0, dtype=np.double)
    # _y_coordinates: np.empty(0, dtype=np.double)
    # _distance: np.empty(0, dtype=np.double)
    # length: float = 0.0
    # branch_offsets: np.empty(0, dtype=np.double)
    # nodes: np.empty(0, dtype=np.double)
    # present: np.empty(0, dtype=bool)

    def __init__(
        self,
        geometry: np.ndarray,
        branch_offsets: np.ndarray = None,
        mask: np.ndarray = None,
    ) -> None:
        # Check that the array has two collumns (x and y)
        assert geometry.shape[1] == 2

        # Split in x and y
        self.geometry = geometry
        self._x_coordinates = geometry[:, 0]
        self._y_coordinates = geometry[:, 1]

        # Calculate distance of coordinates along line
        segment_distances = np.hypot(
            np.diff(self._x_coordinates), np.diff(self._y_coordinates)
        )
        self._distance = np.concatenate([[0], np.cumsum(segment_distances)])
        self.length = segment_distances.sum()

        # Check if mask and branch offsets (if both given) have same shape
        if (
            mask is not None
            and branch_offsets is not None
            and branch_offsets.shape != mask.shape
        ):
            raise ValueError("Mask and branch offset have different shape.")

        # Set branch offsets
        self.branch_offsets = branch_offsets
        # Calculate node positions
        if branch_offsets is not None:
            # if (branch_offsets < 0.0).any():
            #     raise ValueError("Minimum branch offset is smaller than 0.0")
            # elif (branch_offsets > self.length).any():
            #     raise ValueError(
            #         f"Maximum branch offset ({max(branch_offsets)}) is larger than the branch length ({self.length})"
            #     )
            self.node_xy = self.interpolate(branch_offsets)

        # Set which of the nodes are present
        if (mask is None) and (branch_offsets is not None):
            self.mask = np.full(branch_offsets.shape, False)
        else:
            self.mask = mask

    def generate_nodes(
        self,
        mesh1d_edge_length: float,
        seperate_structures: bool = False,
        max_dist_to_struc: float = None,
    ):
        if seperate_structures:
            raise NotImplementedError(
                "Taking into account structure positions is not implemented."
            )

        # Generate offsets
        self.branch_offsets = self._generate_1d_spacing(
            anchor_pts=[0.0, self.length], mesh1d_edge_length=mesh1d_edge_length
        )
        # Calculate node positions
        self.node_xy = self.interpolate(self.branch_offsets)
        # Add mask (all False)
        self.mask = np.full(self.branch_offsets.shape, False)

    def _generate_1d_spacing(self, anchor_pts: List[float], mesh1d_edge_length: float):
        """
        Generates 1d distances, called by function generate offsets
        """
        offsets = []
        for i in range(len(anchor_pts) - 1):
            section_length = anchor_pts[i + 1] - anchor_pts[i]
            if section_length <= 0.0:
                raise ValueError("Section length must be larger than 0.0")
            nnodes = max(2, int(round(section_length / mesh1d_edge_length) + 1))
            offsets.extend(
                np.linspace(
                    anchor_pts[i], anchor_pts[i + 1], nnodes - 1, endpoint=False
                ).tolist()
            )
        offsets.append(anchor_pts[-1])

        return np.asarray(offsets)

    def interpolate(self, distance: Union[float, np.ndarray]) -> np.ndarray:
        """Interpolate coordinates along branch by length
        #TODO: How to handle these kind of union datatypes? The array should also consist of floats
        Args:
            distance (Union[float, np.ndarray]): Length
        """
        intpcoords = np.stack(
            [
                np.interp(distance, self._distance, self._x_coordinates),
                np.interp(distance, self._distance, self._y_coordinates),
            ],
            axis=1,
        )

        return intpcoords


class Link1d2d(BaseModel):

    meshkernel: Optional[mk.MeshKernel] = Field(default_factory=mk.MeshKernel)

    link1d2d_id: np.ndarray = np.empty(0, object)
    link1d2d_long_name: np.ndarray = np.empty(0, object)
    link1d2d_contact_type: np.ndarray = np.empty(0, np.int32)
    link1d2d: np.ndarray = np.empty((0, 2), np.int32)

    def is_empty(self) -> bool:
        return self.link1d2d.size == 0

    def read_file(self, file_path: Path) -> None:
        from hydrolib.core.io.net.reader import UgridReader

        reader = UgridReader(file_path)
        reader.read_link1d2d(self)

    def clear(self) -> None:
        """Remove all saved links from the links administration"""
        self.link1d2d_id = np.empty(0, object)
        self.link1d2d_long_name = np.empty(0, object)
        self.link1d2d_contact_type = np.empty(0, np.int32)
        self.link1d2d = np.empty((0, 2), np.int32)

    def _process(self) -> None:
        """
        Get links from meshkernel and add to the array with link administration
        """
        contacts = self.meshkernel.contacts_get()

        self.link1d2d = np.append(
            self.link1d2d,
            np.stack([contacts.mesh1d_indices, contacts.mesh2d_indices], axis=1),
            axis=0,
        )
        self.link1d2d_contact_type = np.append(
            self.link1d2d_contact_type, np.full(contacts.mesh1d_indices.size, 3)
        )
        self.link1d2d_id = np.append(
            self.link1d2d_id,
            np.array([f"{n1d:d}_{f2d:d}" for n1d, f2d in self.link1d2d]),
        )
        self.link1d2d_long_name = np.append(
            self.link1d2d_long_name,
            np.array([f"{n1d:d}_{f2d:d}" for n1d, f2d in self.link1d2d]),
        )

    def _link_from_1d_to_2d(
        self, node_mask: np.ndarray, polygon: mk.GeometryList = None
    ):
        """Connect 1d nodes to 2d face circumcenters. A list of branchid's can be given
        to indicate where the 1d-side of the connections should be made. A polygon can
        be given to indicate where the 2d-side of the connections should be made.

        Note that the links are added to the already existing links. To remove these, use the method "clear".

        Args:
            node_mask (np.ndarray): Array indicating what 1d nodes should be connected. Defaults to None.
            polygon (mk.GeometryList): Coordinates of the area within which the 2d side of the links are connected.
        """

        # Computes Mesh1d-Mesh2d contacts, where each single Mesh1d node is connected to one Mesh2d face circumcenter.
        # The boundary nodes of Mesh1d (those sharing only one Mesh1d edge) are not connected to any Mesh2d face.
        self.meshkernel.contacts_compute_single(node_mask=node_mask, polygons=polygon)
        self._process()

        # Note that the function "contacts_compute_multiple" also computes the connections, but does not take into account
        # a bounding polygon or the end points of the 1d mesh.
        # self._mk.contacts_compute_multiple(self, node_mask)

    def _link_from_2d_to_1d_intersecting(self):
        raise NotImplementedError()

    def _link_from_2d_to_1d_lateral(self):
        raise NotImplementedError()

        # # Computes Mesh1d-Mesh2d contacts, where a Mesh2d face per polygon is connected to the closest Mesh1d node.
        # self._mk.contacts_compute_with_polygons(self, node_mask, polygons)

        # # Computes Mesh1d-Mesh2d contacts, where Mesh1d nodes are connected to the Mesh2d face mass centers containing the input point.
        # self._mk.contacts_compute_with_points(self, node_mask, points)

        # # Computes Mesh1d-Mesh2d contacts, where Mesh1d nodes are connected to the closest Mesh2d faces at the boundary
        # self._mk.contacts_compute_boundary(self, node_mask, polygons, search_radius)


class Mesh1d(BaseModel):
    """"""

    meshkernel: mk.MeshKernel = Field(default_factory=mk.MeshKernel)

    branches: Dict[str, Branch] = {}

    network1d_node_id: np.ndarray = np.empty(0, object)
    network1d_node_long_name: np.ndarray = np.empty(0, object)
    network1d_node_x: np.ndarray = np.empty(0, np.double)
    network1d_node_y: np.ndarray = np.empty(0, np.double)
    network1d_branch_id: np.ndarray = np.empty(0, object)
    network1d_branch_long_name: np.ndarray = np.empty(0, object)
    network1d_branch_length: np.ndarray = np.empty(0, np.double)
    network1d_branch_order: np.ndarray = np.empty(0, np.int32)
    network1d_edge_nodes: np.ndarray = np.empty((0, 2), np.int32)
    network1d_geom_x: np.ndarray = np.empty(0, np.double)
    network1d_geom_y: np.ndarray = np.empty(0, np.double)
    network1d_part_node_count: np.ndarray = np.empty(0, np.int32)

    mesh1d_node_x: np.ndarray = np.empty(0, np.double)
    mesh1d_node_y: np.ndarray = np.empty(0, np.double)
    mesh1d_node_id: np.ndarray = np.empty(0, object)
    mesh1d_node_long_name: np.ndarray = np.empty(0, object)
    mesh1d_node_branch_id: np.ndarray = np.empty(0, np.int32)
    mesh1d_node_branch_offset: np.ndarray = np.empty(0, np.double)

    mesh1d_edge_nodes: np.ndarray = np.empty((0, 2), np.int32)
    mesh1d_edge_x: np.ndarray = np.empty(0, np.double)
    mesh1d_edge_y: np.ndarray = np.empty(0, np.double)
    mesh1d_edge_branch_id: np.ndarray = np.empty(0, np.int32)
    mesh1d_edge_branch_offset: np.ndarray = np.empty(0, np.double)

    def is_empty(self) -> bool:
        return self.mesh1d_node_x.size == 0

    def _get_mesh1d(self) -> mk.Mesh1d:
        """Return mesh1d from meshkernel. Note that the meshkernel.Mesh1d instance
        does not contain all mesh attributes that are contained in this class"""
        return self.meshkernel.mesh1d_get()

    def _set_mesh1d(self) -> None:
        self.meshkernel.mesh1d_set(
            mk.Mesh1d(
                node_x=self.mesh1d_node_x.astype(np.float64),
                node_y=self.mesh1d_node_y.astype(np.float64),
                edge_nodes=self.mesh1d_edge_nodes.ravel().astype(np.int32),
            )
        )

    def _process_network1d(self) -> None:
        """
        Determine x, y locations of mesh1d nodes based on the network1d
        """
        # Create a list of coordinates to create the branches from
        ngeom = list(zip(self.network1d_geom_x, self.network1d_geom_y))

        self.branches.clear()

        for i, (name, nnodes) in enumerate(
            zip(self.network1d_branch_id, self.network1d_part_node_count)
        ):

            # Create network branch
            # Get geometry of branch from network geometry
            geometry = np.array([ngeom.pop(0) for _ in range(nnodes)])
            # Get branch offsets
            idx = self.mesh1d_node_branch_id == i
            branch_offsets = self.mesh1d_node_branch_offset[idx]
            mask = np.full(branch_offsets.shape, False)

            # Determine if a start or end coordinate needs to be added for constructing a complete branch
            # As nodes are re-used, the last and first branch_offsets are often missing. However, they are still used
            # for determining the length along the discretized branch.
            if branch_offsets.size == 0 or not np.isclose(branch_offsets[0], 0.0):
                branch_offsets = np.concatenate([[0], branch_offsets])
                mask = np.concatenate([[True], mask])
            length = np.hypot(*np.diff(geometry, axis=0).T).sum()
            if not np.isclose(branch_offsets[-1], length):
                branch_offsets = np.concatenate([branch_offsets, [length]])
                mask = np.concatenate([mask, [True]])

            # Create instance of branch object and add to dictionary
            geo_branch = Branch(geometry, branch_offsets=branch_offsets, mask=mask)
            self.branches[name.strip()] = geo_branch

        # Convert list with all coordinates (except the appended ones for the schematized branches) to arrays
        node_x, node_y = np.vstack(
            [branch.node_xy[~branch.mask] for branch in self.branches.values()]
        ).T

        # Add to variables
        self.mesh1d_node_x = node_x
        self.mesh1d_node_y = node_y

        # Calculate edge coordinates
        edge_x, edge_y = np.vstack(
            [
                branch.interpolate(
                    self.mesh1d_edge_branch_offset[self.mesh1d_edge_branch_id == i]
                )
                for i, branch in enumerate(self.branches.values())
            ]
        ).T

        # Add to variables
        self.mesh1d_edge_x = edge_x
        self.mesh1d_edge_y = edge_y

    def _network1d_node_position(self, x: float, y: float) -> Union[np.int32, None]:
        """Determine the position (index) of a x, y coordinate in the network nodes

        Args:
            x (float): x-coordinate
            y (float): y-coordinate

        Returns:
            Union[np.int32, None]: The index of the coordinate. None if not found
        """
        return self._node_position(self.network1d_node_x, self.network1d_node_y, x, y)

    def _mesh1d_node_position(self, x: float, y: float) -> Union[np.int32, None]:
        """Determine the position (index) of a x, y coordinate in the mesh nodes

        Args:
            x (float): x-coordinate
            y (float): y-coordinate

        Returns:
            Union[np.int32, None]: The index of the coordinate. None if not found
        """
        return self._node_position(self.mesh1d_node_x, self.mesh1d_node_y, x, y)

    def _node_position(
        self, arrx: np.ndarray, arry: np.ndarray, x: float, y: float
    ) -> Union[np.int32, None]:
        """Determine the position (index) of a x, y coordinate in a given x and y array

        Args:
            arrx (np.ndarray): x-coordinates in which the position is sought
            arry (np.ndarray): y-coordiantes in which the position is sought
            x (float): x-coordinate to be sought
            y (float): y-coordinate to be sought

        Raises:
            ValueError: If multiple positions are found for the coordinate

        Returns:
            Union[np.int32, None]: The index of the coordinate. None if not found
        """
        pos = np.where(np.isclose(arrx, x) & np.isclose(arry, y))[0]
        if pos.size == 0:
            return None
        elif pos.size == 1:
            return np.int32(pos[0])
        else:
            raise ValueError("Multiple nodes were found at the given position.")

    def _add_branch(
        self,
        branch: Branch,
        name: str = None,
        branch_order: int = -1,
        long_name: str = None,
    ):

        # Check if branch had coordinate discretization
        if branch.branch_offsets.size == 0:
            raise ValueError(
                'Branch has no mesh discretization. Use the function "generate_nodes" solve generate a 1d mesh on the branch.'
            )

        if name in self.network1d_branch_id:
            raise KeyError(f'The branch name "{name}" is already used.')
        if long_name in self.network1d_branch_long_name:
            raise KeyError(f'The branch long name "{long_name}" is already used.')

        branch_nr = len(self.network1d_branch_id)
        if name is None:
            name = f"br{branch_nr:05d}"
        if long_name is None:
            long_name = name

        self.branches[name] = branch

        # Add branch administration
        self.network1d_branch_order = np.append(self.network1d_branch_order, -1)
        self.network1d_branch_length = np.append(
            self.network1d_branch_length, branch.length
        )
        self.network1d_branch_id = np.append(self.network1d_branch_id, name)
        self.network1d_branch_long_name = np.append(
            self.network1d_branch_long_name, long_name
        )

        # Add branch geometry coordinates
        self.network1d_part_node_count = np.append(
            self.network1d_part_node_count, len(branch.geometry)
        )
        self.network1d_geom_x = np.append(self.network1d_geom_x, branch._x_coordinates)
        self.network1d_geom_y = np.append(self.network1d_geom_y, branch._y_coordinates)

        # Network edge node administration
        # -------------------------------
        first_point = branch.geometry[0]
        last_point = branch.geometry[-1]

        # Get offsets from dictionary
        offsets = branch.branch_offsets[:]
        # The number of links on the branch
        nlinks = len(offsets) - 1

        # Check if the first and last point of the branch are already in the set
        first_present = self._network1d_node_position(*first_point) is not None
        if first_present:
            # If present, remove from branch offsets
            offsets = offsets[1:]
            branch.mask[0] = True
        else:
            # If not present, add to network nodes
            self.network1d_node_x = np.append(self.network1d_node_x, first_point[0])
            self.network1d_node_y = np.append(self.network1d_node_y, first_point[1])

            self.network1d_node_id = np.append(
                self.network1d_node_id, "{:.0f}_{:.0f}".format(*first_point)
            )
            self.network1d_node_long_name = np.append(
                self.network1d_node_long_name, "x={:.0f}_y={:.0f}".format(*first_point)
            )

        last_present = self._network1d_node_position(*last_point) is not None
        if last_present:
            # If present, remove from branch offsets
            offsets = offsets[:-1]
            branch.mask[-1] = True
        else:
            # If not present, add to network nodes
            self.network1d_node_x = np.append(self.network1d_node_x, last_point[0])
            self.network1d_node_y = np.append(self.network1d_node_y, last_point[1])

            self.network1d_node_id = np.append(
                self.network1d_node_id, "{:.0f}_{:.0f}".format(*last_point)
            )
            self.network1d_node_long_name = np.append(
                self.network1d_node_long_name, "x={:.0f}_y={:.0f}".format(*last_point)
            )

        # If no points remain, add an extra halfway: each branch should have at least 1 node
        if len(offsets) == 0:
            offsets = np.array([branch.length / 2.0])
            nlinks += 1

        # Get the index of the first and last node, add as edge_nodes
        i_from = self._network1d_node_position(first_point[0], first_point[1])
        i_to = self._network1d_node_position(last_point[0], last_point[1])
        if i_from == i_to:
            raise ValueError(
                "Start and end node are the same. Ring geometries are not accepted."
            )

        self.network1d_edge_nodes = np.append(
            self.network1d_edge_nodes,
            np.array([[i_from, i_to]], dtype=np.int32),
            axis=0,
        )

        # Mesh1d edge node administration
        # -------------------------------
        # First determine the start index. This is equal to the number of already present points
        start_index = len(self.mesh1d_node_branch_id)
        # For each link, create a new edge node connection
        # If the first node is already present, subtract 1, since the first number will be substitud with the present node
        if first_present:
            start_index -= 1
        new_edge_nodes = (
            np.stack([np.arange(nlinks), np.arange(nlinks) + 1], axis=1) + start_index
        ).astype(np.int32)

        # If the first node is present, change the first point of the first edge to the existing point
        if first_present:
            new_edge_nodes[0, 0] = self._mesh1d_node_position(*first_point)
        # If the last node is present, change the last point of the last edge too
        if last_present:
            new_edge_nodes[-1, 1] = self._mesh1d_node_position(*last_point)

        # Add to variables
        self.mesh1d_node_x = np.append(
            self.mesh1d_node_x, branch.node_xy[~branch.mask, 0]
        )
        self.mesh1d_node_y = np.append(
            self.mesh1d_node_y, branch.node_xy[~branch.mask, 1]
        )

        # Add to edge_nodes
        self.mesh1d_edge_nodes = np.append(
            self.mesh1d_edge_nodes, new_edge_nodes, axis=0
        )
        edge_coords = np.stack([self.mesh1d_node_x, self.mesh1d_node_y], axis=1)[
            new_edge_nodes
        ].mean(1)
        edge_offsets = (branch.branch_offsets[:-1] + branch.branch_offsets[1:]) / 2

        self.mesh1d_edge_branch_id = np.append(
            self.mesh1d_edge_branch_id, np.full(len(edge_coords), branch_nr)
        )
        self.mesh1d_edge_branch_offset = np.append(
            self.mesh1d_edge_branch_offset, edge_offsets
        )

        self.mesh1d_edge_x = np.append(self.mesh1d_edge_x, edge_coords[:, 0])
        self.mesh1d_edge_y = np.append(self.mesh1d_edge_y, edge_coords[:, 1])

        # Update names of nodes
        mesh_point_names = np.array(
            [f"{name}_{offset:.2f}" for offset in offsets], dtype=object
        )
        self.mesh1d_node_id = np.append(self.mesh1d_node_id, mesh_point_names)
        self.mesh1d_node_long_name = np.append(
            self.mesh1d_node_long_name, mesh_point_names
        )

        # Add mesh1d nodes
        self.mesh1d_node_branch_id = np.append(
            self.mesh1d_node_branch_id, np.full(len(offsets), branch_nr)
        )
        self.mesh1d_node_branch_offset = np.append(
            self.mesh1d_node_branch_offset, offsets
        )

        # self._process_edges_for_branch()

    def _process_edges_for_branch(self, branch_id: str) -> None:

        branch = self.branches[branch_id]

        edge_coords = (branch.node_xy[:-1] + branch.node_xy[1:]) / 2.0
        edge_offsets = (branch.branch_offsets[:-1] + branch.branch_offsets[1:]) / 2

        self.mesh1d_edge_branch_id = np.append(
            self.mesh1d_edge_branch_id, np.full(len(edge_coords), i)
        )
        self.mesh1d_edge_branch_offset = np.append(
            self.mesh1d_edge_branch_offset, edge_offsets
        )
        self.mesh1d_edge_x = np.append(self.mesh1d_edge_x, edge_coords[:, 0])
        self.mesh1d_edge_y = np.append(self.mesh1d_edge_y, edge_coords[:, 1])

    def get_node_mask(self, branchids: List[str] = None):
        """Get node mask, give a mask with True for each node that is in the given branchid list"""

        mask = np.full(self.mesh1d_node_id.shape, False, dtype=bool)
        if branchids is None:
            mask[:] = True
            return mask

        # Get number (index) of given branches
        idx = np.where(np.isin(self.network1d_branch_id, branchids))[0]
        if idx.size == 0:
            raise KeyError("No branches corresponding to the given keys were found.")

        mask[np.isin(self.mesh1d_node_branch_id, idx)] = True

        return mask


class Network:

    """"""

    def __init__(self) -> None:
        self.meshkernel = mk.MeshKernel()

        self._mesh1d = Mesh1d(meshkernel=self.meshkernel)
        self._mesh2d = Mesh2d(meshkernel=self.meshkernel)
        self._link1d2d = Link1d2d(meshkernel=self.meshkernel)

        # Spatial index (rtree)
        # self._idx = index.Index()

    @classmethod
    def from_file(cls, file_path: Path) -> Network:
        """Read network from file. This classmethod checks what mesh components (mesh1d & network1d, mesh2d, link1d2d) are
        present, and loads them one by one.

        Args:
            file (Path): path to netcdf file with network data

        Returns:
            Network: The instance of the class itself that is returned
        """
        from hydrolib.core.io.net.reader import UgridReader

        network = cls()
        ds = nc.Dataset(file_path)  # type: ignore[import]

        reader = UgridReader(file_path)

        if reader._explorer.mesh1d_key is not None:
            reader.read_mesh1d_network1d(network._mesh1d)

        if reader._explorer.mesh2d_key is not None:
            reader.read_mesh2d(network._mesh2d)

        if reader._explorer.link1d2d_key is not None:
            reader.read_link1d2d(network._link1d2d)

        ds.close()

        return network

    def to_file(self, file: Path) -> None:
        """Write network to file

        Args:
            file (Path): File where _net.nc is written to.
        """
        from hydrolib.core.io.net.writer import UgridWriter

        writer = UgridWriter()
        writer.write(self, file)

    def link1d2d_from_1d_to_2d(
        self, branchids: List[str] = None, polygon: GeometryList = None
    ) -> None:
        self._mesh1d._set_mesh1d()
        self._mesh2d._set_mesh2d()

        node_mask = self._mesh1d.get_node_mask(branchids)
        if polygon is None:
            polygon = self.meshkernel.mesh2d_get_mesh_boundaries_as_polygons()

        self._link1d2d._link_from_1d_to_2d(node_mask, polygon=polygon)

    def mesh2d_create_rectilinear_within_bounds(
        self, extent: tuple, dx: float, dy: float
    ) -> None:
        self._mesh2d.create_rectilinear(extent=extent, dx=dx, dy=dy)

    def mesh2d_create_triangular_within_polygon(self, polygon: mk.GeometryList) -> None:
        raise NotImplementedError()

    def mesh2d_clip_mesh(
        self, polygon: mk.GeometryList, deletemeshoption: int = 1
    ) -> None:
        self._mesh2d.clip(polygon=polygon, deletemeshoption=deletemeshoption)

    def mesh2d_refine_mesh(self, polygon: mk.GeometryList, level: int = 1) -> None:
        self._mesh2d.refine(polygon=polygon, level=level)

    def mesh1d_add_branch(
        self,
        branch: Branch,
        name: str = None,
        branch_order: int = -1,
        long_name: str = None,
    ) -> None:
        self._mesh1d._add_branch(
            branch=branch, name=name, branch_order=branch_order, long_name=long_name
        )


class NetworkModel(FileModel):
    """Network model representation."""

    class Config:
        extra = "allow"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # We're keeping the `Network` class completely outside Pydantic
        if self.filepath and self.filepath.is_file():
            self.network = Network.from_file(self.filepath)
        else:
            self.network = Network()

    @property
    def _mesh1d(self):
        return self.network._mesh1d

    @property
    def _mesh2d(self):
        return self.network._mesh2d

    @property
    def _link1d2d(self):
        return self.network._link1d2d

    @classmethod
    def _ext(cls) -> str:
        return ".nc"

    @classmethod
    def _filename(cls) -> str:
        return "network"

    def _save(self, folder):
        filename = Path(self.filepath.name) if self.filepath else self._generate_name()
        self.filepath = folder / filename
        self.network.to_file(self.filepath)

    def _parse(self, _):
        return {}

    @classmethod
    def _get_serializer(cls):
        # Unused, but requires abstract implementation
        pass

    @classmethod
    def _get_parser(cls):
        # Unused, but requires abstract implementation
        pass



from types import new_class
from typing import Any, Callable, Generator, Generic, List, TypeVar, cast


def example(a: float, b: float = 1.0) -> float:
    """[summary]

    Args:
        a (float): [description]
        b (float): [description]

    Returns:
        float: [description]
    """
    return a * b


def to_key(string: str) -> str:
    return string.lower().replace(" ", "_").replace("-", "")


def to_list(item: Any) -> List[Any]:
    """Puts the specified item in a list if it is an instance of `dict`.

    Attributes:
        item: The item to put in a list.

    Returns:
        List: A list with the specified item.
    """

    if not isinstance(item, list):
        return [item]
    return item



from abc import ABC, abstractclassmethod
from datetime import datetime
from pathlib import Path
from typing import Callable, List, Literal, Optional, Type, Union

from pydantic import Field, validator

from hydrolib.core import __version__
from hydrolib.core.basemodel import BaseModel, FileModel
from hydrolib.core.io.dimr.parser import DIMRParser
from hydrolib.core.io.dimr.serializer import DIMRSerializer
from hydrolib.core.io.mdu.models import FMModel
from hydrolib.core.utils import to_list


class KeyValuePair(BaseModel):
    """Key value pair to specify settings and parameters.

    Attributes:
        key: The key.
        value: The value.
    """

    key: str
    value: str


class Component(BaseModel, ABC):
    """
    Specification of a BMI-compliant model component instance that will be executed by DIMR.

    Attributes:
        library: The library name of the compoment.
        name: The component name.
        workingDir: The working directory.
        inputFile: The name of the input file.
        process: Number of subprocesses in the component.
        setting: A list of variables that are provided to the BMI model before initialization.
        parameter: A list of variables that are provided to the BMI model after initialization.
        mpiCommunicator: The MPI communicator value.
        model: The model represented by this component.
    """

    library: str
    name: str
    workingDir: Path
    inputFile: Path
    process: Optional[int]
    setting: Optional[List[KeyValuePair]] = []
    parameter: Optional[List[KeyValuePair]] = []
    mpiCommunicator: Optional[str]

    model: Optional[FileModel]

    @property
    def filepath(self):
        return self.workingDir / self.inputFile

    @abstractclassmethod
    def get_model(cls) -> Type[FileModel]:
        raise NotImplementedError("Model not implemented yet.")

    @validator("setting", "parameter", pre=True)
    def validate_setting(cls, v):
        return to_list(v)

    def is_intermediate_link(self) -> bool:
        return True


class FMComponent(Component):
    library: Literal["dflowfm"]

    @classmethod
    def get_model(cls):
        return FMModel


class RRComponent(Component):
    library: Literal["rr_dll"]

    @classmethod
    def get_model(cls):
        raise NotImplementedError("Model not implemented yet.")


class Documentation(BaseModel):
    """
    Information on the present DIMR configuration file.

    Attributes:
        fileVersion: The DIMR file version.
        createdBy: Creators of the DIMR file.
        creationDate: The creation date of the DIMR file.
    """

    fileVersion: str = "1.3"
    createdBy: str = f"hydrolib-core {__version__}"
    creationDate: datetime = Field(default_factory=datetime.utcnow)


class GlobalSettings(BaseModel):
    """
    Global settings for the DIMR configuration.

    Attributes:
        logger_ncFormat: NetCDF format type for logging.
    """

    logger_ncFormat: int


class ComponentOrCouplerRef(BaseModel):
    """
    Reference to a BMI-compliant model component instance.

    Attributes:
        name: Name of the reference to a BMI-compliant model component instance.
    """

    name: str


class CoupledItem(BaseModel):
    """
    Specification of an item that has to be exchanged.

    Attributes:
        sourceName: Name of the item at the source component.
        targetName: Name of the item at the target component.
    """

    sourceName: str
    targetName: str

    def is_intermediate_link(self) -> bool:
        # TODO set to True once we replace Paths with FileModels
        return False


class Logger(BaseModel):
    """
    Used to log values to the specified file in workingdir for each timestep

    Attributes:
        workingDir: Directory where the log file is written.
        outputFile: Name of the log file.
    """

    workingDir: Path
    outputFile: Path


class Coupler(BaseModel):
    """
    Specification of the coupling actions to be performed between two BMI-compliant model components.

    Attributes:
        name: The name of the coupler.
        sourceComponent: The component that provides the data to has to be exchanged.
        targetComponent: The component that consumes the data to has to be exchanged.
        item: A list of items that have to be exchanged.
        logger: Logger for logging the values that get exchanged.
    """

    name: str
    sourceComponent: str
    targetComponent: str
    item: List[CoupledItem] = []
    logger: Optional[Logger]

    @validator("item", pre=True)
    def validate_item(cls, v):
        return to_list(v)

    def is_intermediate_link(self) -> bool:
        # TODO set to True once we replace Paths with FileModels
        return False


class StartGroup(BaseModel):
    """
    Specification of model components and couplers to be executed with a certain frequency.

    Attributes:
        time: Time frame specification for the present group: start time, stop time and frequency.
              Expressed in terms of the time frame of the main component.
        start: Ordered list of components to be executed.
        coupler: Oredered list of couplers to be executed.
    """

    time: str
    start: List[ComponentOrCouplerRef] = []
    coupler: List[ComponentOrCouplerRef] = []

    @validator("start", "coupler", pre=True)
    def validate_start(cls, v):
        return to_list(v)


class Parallel(BaseModel):
    """
    Specification of a parallel control flow: one main component and a group of related components and couplers.
    Step wise execution order according to order in parallel control flow.

    Attributes:
        startGroup: Group of components and couplers to be executed.
        start: Main component to be executed step wise (provides start time, end time and time step).
    """

    startGroup: StartGroup
    start: ComponentOrCouplerRef


class Control(BaseModel):
    """
    Control flow specification for the DIMR-execution.

    Attributes:
        parallel: Specification of a control flow that has to be executed in parallel.
        start: Reference to the component instance to be started.
    """

    parallel: Optional[List[Parallel]] = []
    start: Optional[List[ComponentOrCouplerRef]] = []

    @validator("parallel", "start", pre=True)
    def validate_parallel(cls, v):
        return to_list(v)

    def is_intermediate_link(self) -> bool:
        # TODO set to True once we replace Paths with FileModels
        return False


class DIMR(FileModel):
    """DIMR model representation."""

    documentation: Documentation = Documentation()
    control: Control = Control()
    component: List[Union[RRComponent, FMComponent, Component]] = []
    coupler: Optional[List[Coupler]] = []
    waitFile: Optional[str]
    global_settings: Optional[GlobalSettings]

    @validator("component", "coupler", pre=True)
    def validate_component(cls, v):
        return to_list(v)

    def dict(self, *args, **kwargs):
        """Converts this object recursively to a dictionary.

        Returns:
            dict: The created dictionary for this object.
        """
        return self._to_serializable_dict(self)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # After initilization, try to load all component models
        if self.filepath:
            for comp in self.component:
                fn = self.filepath.parent / comp.filepath
                try:
                    comp.model = comp.get_model()(filepath=fn)
                except NotImplementedError:
                    continue

    @classmethod
    def _ext(cls) -> str:
        return ".xml"

    @classmethod
    def _filename(cls) -> str:
        return "dimrconfig"

    @classmethod
    def _get_serializer(cls) -> Callable:
        return DIMRSerializer.serialize

    @classmethod
    def _get_parser(cls) -> Callable:
        return DIMRParser.parse

    def _to_serializable_dict(self, obj) -> dict:
        if not hasattr(obj, "__dict__"):
            return obj

        result = {}

        for key, val in obj.__dict__.items():
            if (
                key.startswith("_")
                or key == "filepath"
                or isinstance(val, FileModel)
                or val is None
            ):
                continue

            element = []
            if isinstance(val, list):
                for item in val:
                    element.append(self._to_serializable_dict(item))
            else:
                element = self._to_serializable_dict(val)
            result[key] = element

        return result
